module frame_capture (
        input clk,
        input rst,

        // Image input signals
        input new_frame,
        input new_line,
        input new_pixel,
        input [15:0] pixel,

        // SDRAM interface
        output [22:0] addr,
        output rw,
        output [31:0] data_in,
        input [31:0] data_out,
        input busy,
        output in_valid,
        input out_valid,

        // User interface
        input capture_image,
        output image_captured,

        output [7:0] led,
        output laser
    );

    localparam STATE_SIZE = 4;
    localparam IDLE = 0,
               WAIT_FRAME = 1,
               WAIT_PIXEL_1 = 2,
               WAIT_PIXEL_2 = 3,
               WRITE_SDRAM = 4,
               INC_ADDR = 5,
               READ_PIXEL_1 = 6,
               WAIT_RAM_1 = 7,
               READ_PIXEL_2 = 8,
               WAIT_RAM_2 = 9,
               READ_PIXEL_3 = 10,
               WAIT_RAM_3 = 11,
               READ_PIXEL_4 = 12,
               WAIT_RAM_4 = 13,
               WRITE_PIXEL = 14,
               WAIT_WRITE = 15;

    localparam PIXEL_COUNT = 1600*1200;

    reg [STATE_SIZE-1:0] state_d, state_q;

    reg [22:0] addr_d, addr_q;
    reg rw_d, rw_q;
    reg [31:0] data_d, data_q;
    reg in_valid_d, in_valid_q;
    reg [10:0] pixel_d, pixel_q;
    reg [10:0] row_d, row_q;
    reg [1:0] frame_d, frame_q;
    reg image_captured_d, image_captured_q;

    reg [7:0] led_d, led_q;
    reg laser_d, laser_q;

    reg [18:0] img_pixel_d, img_pixel_q;
    reg [31:0] p2_d, p2_q, p1_d, p1_q, p3_d, p3_q, p4_d, p4_q;

    reg [5:0] r1,g1,b1,r2,g2,b2;
    reg [6:0] s1,s2;

    assign data_in = data_q;
    assign addr = addr_q;
    assign rw = rw_q;
    assign in_valid = in_valid_q;
    assign led = led_q;
    assign laser = laser_q;
    assign image_captured = image_captured_q;

    always @(*) begin
        state_d = state_q;
        addr_d = addr_q;
        rw_d = rw_q;
        data_d = data_q;
        in_valid_d = 1'b0;
        pixel_d = pixel_q;
        row_d = row_q;
        frame_d = frame_q;
        image_captured_d = image_captured_q;

        led_d = state_q;
        laser_d = laser_q;

        img_pixel_d = img_pixel_q;
        p1_d = p1_q;
        p2_d = p2_q;
        p3_d = p3_q;
        p4_d = p4_q;

        if (new_pixel)
            pixel_d = pixel_q + 1'b1;
        if (new_line) begin
            row_d = row_q + 1'b1;
            if (new_pixel)
                pixel_d = 11'd1;
            else
                pixel_d = 11'd0;
        end

        if (new_frame) begin
            row_d = -11'd1;
            pixel_d = 11'd0;
            frame_d = frame_q + 1'b1;
        end

        case (state_q)
            IDLE: begin
                laser_d = 1'b1;
                addr_d = 23'd0;
                frame_d = 1'b0;
                if (capture_image) begin
                    state_d = WAIT_FRAME;
                    image_captured_d = 1'b0;
                end
            end
            WAIT_FRAME: begin
                if (new_frame) begin
                    frame_d = 1'b0;
                    if (new_pixel) begin
                        data_d[15:0] = pixel;
                        state_d = WAIT_PIXEL_2;
                    end else
                        state_d = WAIT_PIXEL_1;
                end

            end
            WAIT_PIXEL_1: begin
                if (new_pixel) begin
                    //data_d [15:0] = addr_q[15:0];
                    data_d[15:0] = pixel;
                    state_d = WAIT_PIXEL_2;
                end
            end
            WAIT_PIXEL_2: begin
                if (new_pixel) begin
                    //data_d [31:16] = addr_q[22:16];
                    data_d[31:16] = pixel;
                    state_d = WRITE_SDRAM;
                end
            end
            WRITE_SDRAM: begin
                if (!busy) begin
                    rw_d = 1'b1; //write
                    in_valid_d = 1'b1;
                    state_d = INC_ADDR;
                end
            end
            INC_ADDR: begin
                addr_d = (row_q * 800) + pixel_q[10:1] + (frame_q[0] * 960000);
                state_d = WAIT_PIXEL_1;

                if (row_q > 400 && !frame_q[0])
                    laser_d = 1'b0; // turn on/off laser for next frame

                if (frame_q == 2'd2) begin
                    state_d = READ_PIXEL_1;
                    img_pixel_d = 20'd0;
                end
            end
            READ_PIXEL_1: begin
                if (!busy) begin
                    addr_d = {img_pixel_q, 1'b0};
                    rw_d = 1'b0; //read
                    in_valid_d = 1'b1;
                    state_d = WAIT_RAM_1;
                end
            end
            WAIT_RAM_1: begin
                if (out_valid) begin
                    p1_d = data_out;
                    state_d = READ_PIXEL_2;
                end
            end
            READ_PIXEL_2: begin
                if (!busy) begin
                    addr_d = {img_pixel_q, 1'b1};
                    rw_d = 1'b0; //read
                    in_valid_d = 1'b1;
                    state_d = WAIT_RAM_2;
                end
            end
            WAIT_RAM_2: begin
                if (out_valid) begin
                    p2_d = data_out;
                    state_d = READ_PIXEL_3;
                end
            end
            READ_PIXEL_3: begin
                if (!busy) begin
                    addr_d = {img_pixel_q, 1'b0} + 960000;
                    rw_d = 1'b0; //read
                    in_valid_d = 1'b1;
                    state_d = WAIT_RAM_3;
                end
            end
            WAIT_RAM_3: begin
                if (out_valid) begin
                    p3_d = data_out;
                    state_d = READ_PIXEL_4;
                end
            end
            READ_PIXEL_4: begin
                if (!busy) begin
                    addr_d = {img_pixel_q, 1'b1} + 960000;
                    rw_d = 1'b0; //read
                    in_valid_d = 1'b1;
                    state_d = WAIT_RAM_4;
                end
            end
            WAIT_RAM_4: begin
                if (out_valid) begin
                    p4_d = data_out;
                    state_d = WRITE_PIXEL;
                end
            end
            WRITE_PIXEL: begin
                if (!busy) begin
                    r1 = {p1_q[15:11], p1_q[15]};
                    g1 = p1_q[10:5];
                    b1 = {p1_q[4:0], p1_q[4]};
                    r2 = {p3_q[15:11], p3_q[15]};
                    g2 = p3_q[10:5];
                    b2 = {p3_q[4:0], p3_q[4]};
                    r1 = r1 > r2 ? r1 - r2 : 6'd0;
                    g1 = g1 > g2 ? g1 - g2 : 6'd0;
                    b1 = b1 > b2 ? b1 - b2 : 6'd0;
                    s1 = g1 + b1;
                    data_d[7:0] = r1 > s1 ? r1 - s1 : 7'd0;

                    r1 = {p1_q[31:27], p1_q[31]};
                    g1 = p1_q[26:21];
                    b1 = {p1_q[20:16], p1_q[20]};
                    r2 = {p3_q[31:27], p3_q[31]};
                    g2 = p3_q[26:21];
                    b2 = {p3_q[20:16], p3_q[20]};
                    r1 = r1 > r2 ? r1 - r2 : 6'd0;
                    g1 = g1 > g2 ? g1 - g2 : 6'd0;
                    b1 = b1 > b2 ? b1 - b2 : 6'd0;
                    s1 = g1 + b1;
                    data_d[15:8] = r1 > s1 ? r1 - s1 : 7'd0;

                    r1 = {p2_q[15:11], p2_q[15]};
                    g1 = p2_q[10:5];
                    b1 = {p2_q[4:0], p2_q[4]};
                    r2 = {p4_q[15:11], p4_q[15]};
                    g2 = p4_q[10:5];
                    b2 = {p4_q[4:0], p4_q[4]};
                    r1 = r1 > r2 ? r1 - r2 : 6'd0;
                    g1 = g1 > g2 ? g1 - g2 : 6'd0;
                    b1 = b1 > b2 ? b1 - b2 : 6'd0;
                    s1 = g1 + b1;
                    data_d[23:16] = r1 > s1 ? r1 - s1 : 7'd0;

                    r1 = {p2_q[31:27], p2_q[31]};
                    g1 = p2_q[26:21];
                    b1 = {p2_q[20:16], p2_q[20]};
                    r2 = {p4_q[31:27], p4_q[31]};
                    g2 = p4_q[26:21];
                    b2 = {p4_q[20:16], p4_q[20]};
                    r1 = r1 > r2 ? r1 - r2 : 6'd0;
                    g1 = g1 > g2 ? g1 - g2 : 6'd0;
                    b1 = b1 > b2 ? b1 - b2 : 6'd0;
                    s1 = g1 + b1;
                    data_d[31:24] = r1 > s1 ? r1 - s1 : 7'd0;
                    
                    addr_d = img_pixel_q + 1920000;
                    in_valid_d = 1'b1;
                    rw_d = 1'b1; //write

                    state_d = WAIT_WRITE;

                    img_pixel_d = img_pixel_q + 1'b1;
                    if (img_pixel_q == 479999) begin
                        state_d = IDLE;
                        image_captured_d = 1'b1;
                    end
                end
            end
            WAIT_WRITE: begin
                state_d = READ_PIXEL_1;
            end
            default: state_d = IDLE;
        endcase
    end

    always @(posedge clk) begin
        if (rst) begin
            state_q <= IDLE;
        end else begin
            state_q <= state_d;
        end

        addr_q <= addr_d;
        pixel_q <= pixel_d;
        row_q <= row_d;
        frame_q <= frame_d;
        rw_q <= rw_d;
        data_q <= data_d;
        in_valid_q <= in_valid_d;
        led_q <= led_d;
        laser_q <= laser_d;
        image_captured_q <= image_captured_d;
        img_pixel_q <= img_pixel_d;
        p1_q <= p1_d;
        p2_q <= p2_d;
        p3_q <= p3_d;
        p4_q <= p4_d;
    end

endmodule