module mojo_top(
        input clk,
        input rst_n,
        input cclk,
        output[7:0] led,
        output spi_miso,
        input spi_ss,
        input spi_mosi,
        input spi_sck,
        output [3:0] avr_flags,
        input avr_tx,
        output avr_rx,
        input avr_rx_busy,
        output sdram_clk,
        output sdram_cle,
        output sdram_dqm,
        output sdram_cs,
        output sdram_we,
        output sdram_cas,
        output sdram_ras,
        output [1:0] sdram_ba,
        output [12:0] sdram_a,
        inout [7:0] sdram_dq,

        output sccb_scl,
        inout sccb_sda,
        output camera_xclk,
        input camera_pclk,
        input camera_href,
        input camera_vsync,
        output camera_rst,
        input [7:0] camera_data,
        output laser,
        output [7:0] servo_a,
        output [7:0] servo_b,
        output [7:0] servo_c,
        output [2:0] servo_en
    );

    wire fclk;
    wire rst = ~rst_n;

    wire [22:0] addr;
    wire rw;
    wire [31:0] data_in, data_out;
    wire busy;
    wire in_valid, out_valid;

    localparam RAM_DEVICES = 2;

    wire [22:0] bus_addr [RAM_DEVICES-1:0];
    wire [RAM_DEVICES-1:0] bus_rw;
    wire [31:0] bus_data_in [RAM_DEVICES-1:0], bus_data_out [RAM_DEVICES-1:0];
    wire [RAM_DEVICES-1:0] bus_busy;
    wire [RAM_DEVICES-1:0] bus_in_valid, bus_out_valid;

    wire new_frame, new_line, new_pixel;
    wire [15:0] pixel;
    wire capture_image, image_captured;

    wire [7:0] tx_data, rx_data;
    wire new_tx_data, new_rx_data, tx_busy;

    wire [23:0] pwm;
    wire [4:0] servo_select;
    wire [15:0] servo_position;
    wire servo_update;

    wire [5:0] reg_addr;
    wire reg_write, reg_new_req;
    wire [7:0] reg_write_value, reg_read_value;

    assign servo_en = 3'b111;
    assign servo_a = pwm[7:0];
    assign servo_b = pwm[15:8];
    assign servo_c = pwm[23:16];

    assign led[7:6] = bus_busy;

    sdram_clk_gen clkram_clk_gen (
            .clk_in(clk),
            .clk_out(fclk)
        );

    avr_interface #(.CLK_RATE(100000000), .SERIAL_BAUD_RATE(500000)) avr_interface (
            .clk(fclk),
            .rst(rst),
            .cclk(cclk),
            .spi_miso(spi_miso),
            .spi_mosi(spi_mosi),
            .spi_sck(spi_sck),
            .spi_ss(spi_ss),
            .avr_flags(avr_flags),
            .tx(avr_rx),
            .rx(avr_tx),
            .reg_addr(reg_addr),
            .write(reg_write),
            .new_req(reg_new_req),
            .write_value(reg_write_value),
            .read_value(reg_read_value)
        );

    reg_ctrl reg_ctrl (
            .clk(fclk),
            .rst(rst),
            .reg_addr(reg_addr),
            .write(reg_write),
            .new_req(reg_new_req),
            .write_value(reg_write_value),
            .read_value(reg_read_value),
            .addr(bus_addr[1]),
            .rw(bus_rw[1]),
            .data_in(bus_data_in[1]),
            .data_out(bus_data_out[1]),
            .busy(bus_busy[1]),
            .in_valid(bus_in_valid[1]),
            .out_valid(bus_out_valid[1]),
            .capture_image(capture_image),
            .image_captured(image_captured),
            .servo_select(servo_select),
            .servo_position(servo_position),
            .servo_update(servo_update)
        );

    sdram sdram (
            .clk(fclk),
            .rst(rst),
            .sdram_clk(sdram_clk),
            .sdram_cle(sdram_cle),
            .sdram_cs(sdram_cs),
            .sdram_cas(sdram_cas),
            .sdram_ras(sdram_ras),
            .sdram_we(sdram_we),
            .sdram_dqm(sdram_dqm),
            .sdram_ba(sdram_ba),
            .sdram_a(sdram_a),
            .sdram_dq(sdram_dq),
            .addr(addr),
            .rw(rw),
            .data_in(data_in),
            .data_out(data_out),
            .busy(busy),
            .in_valid(in_valid),
            .out_valid(out_valid)
        );

    ram_bus #(.DEVICES(RAM_DEVICES)) ram_bus (
            .clk(fclk),
            .rst(rst),
            .addr(addr),
            .rw(rw),
            .data_in(data_in),
            .data_out(data_out),
            .busy(busy),
            .in_valid(in_valid),
            .out_valid(out_valid),
            .bus_addr({bus_addr[1],bus_addr[0]}),
            .bus_rw(bus_rw),
            .bus_data_in({bus_data_in[1],bus_data_in[0]}),
            .bus_data_out({bus_data_out[1],bus_data_out[0]}),
            .bus_busy(bus_busy),
            .bus_in_valid(bus_in_valid),
            .bus_out_valid(bus_out_valid),
            .act(led[5:4])
        );

    ov2640 camera (
            .clk(fclk),
            .rst(rst),
            .scl(sccb_scl),
            .sda(sccb_sda),
            .xclk(camera_xclk),
            .pclk(camera_pclk),
            .href(camera_href),
            .vsync(camera_vsync),
            .data(camera_data),
            .rst_cm(camera_rst),
            .new_frame(new_frame),
            .new_line(new_line),
            .new_pixel(new_pixel),
            .pixel(pixel)
        );

    frame_capture frame_capture (
            .clk(fclk),
            .rst(rst),
            .new_frame(new_frame),
            .new_line(new_line),
            .new_pixel(new_pixel),
            .pixel(pixel),
            .addr(bus_addr[0]),
            .rw(bus_rw[0]),
            .data_in(bus_data_in[0]),
            .data_out(bus_data_out[0]),
            .busy(bus_busy[0]),
            .in_valid(bus_in_valid[0]),
            .out_valid(bus_out_valid[0]),
            .capture_image(capture_image),
            .image_captured(image_captured),
            .led(led[3:0]),
            .laser(laser)
        );

    servo_controller servo_controller (
            .clk(fclk),
            .rst(rst),
            .select(servo_select),
            .position(servo_position),
            .in_valid(servo_update),
            .pwm(pwm)
        );
endmodule